import logging

from flask_restful import Resource, abort, request
from marshmallow import Schema, fields
from werkzeug.exceptions import BadRequest

from models.client_model import ClientModel

logging.basicConfig(level=logging.INFO)
logging.getLogger("pika").setLevel(logging.ERROR)


class ClientSchema(Schema):
    client_id = fields.Str(required=True)
    client_name = fields.Str(required=True)
    subscription = fields.Str(required=True)


# Client
class Client(Resource):
    clients_file = None

    def __init__(self):
        self.schema = ClientSchema()
        self.clients_file = ClientModel()

    def get(self, client_id):
        client = self.clients_file.find(client_id)
        # remove secret
        # secret is only returned in initial post response
        del client["secret"]
        if not client:
            abort(404, message="No client with id: {}".format(client_id))
        return client

    def delete(self, client_id):
        client = self.clients_file.find(client_id)
        if not client:
            abort(404, message="No client with id: {}".format(client_id))
        else:
            self.clients_file.remove(client)
        return client, 204

    def put(self, client_id):
        try:
            args = request.get_json()
        except BadRequest:
            abort(400, message="POSTed body is invalid JSON")

        errors = self.schema.validate(args)
        if errors:
            abort(400, message=str(errors))

        client = self.clients_file.find(client_id)
        if not client:
            abort(404, message="No client with id: {}".format(client_id))
        else:
            client = self.clients_file.update(args)
        return client, 201


# ClientList
class ClientList(Resource):
    def __init__(self):
        self.schema = ClientSchema()
        self.clients_file = ClientModel()

    def get(self):
        # return list of clients with secret property removed
        return {
            client_id: (client, client.pop("secret", None))[0]
            for client_id, client in self.clients_file.get().items()
        }

    def post(self):
        try:
            args = request.get_json()
        except BadRequest:
            abort(400, message="POSTed body is invalid JSON")

        errors = self.schema.validate(args)
        if errors:
            abort(400, message=str(errors))

        client = self.clients_file.find(args["client_id"])
        if client:
            abort(409, message="Duplicate client id: {}".format(args["client_id"]))
        else:
            client = self.clients_file.add(args)
        return client, 201
