#!/usr/bin/env python

# Queues
# (per client)
# [client_id]-inbox - receive subscriptions
# [client_id]-outbox - send messages to the bus
# [client-id]-broadcast - send message to all subscribers? - eg notify of downtime
# soar-publisher - fan-in from client-outboxes
# soar-dlq - undeliverables
# soar-broadcast - admin messages forwarded to all client-inboxes regardless of subscriptions
import logging
import os
import socket
import threading
import time

import pika
from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer

from logger import setup_logging
from models.client_model import ClientModel
from rmq import broadcast, forward, publish, subscribe

setup_logging()

RUNNING_CLIENTS = []
CONSUMER_TAGS = {}
EXCHANGES = {
    "publish": "soar_publish",
    "broadcast": "soar_broadcast",
}


class ConfigHandler(FileSystemEventHandler):
    def __init__(self, channel):
        self.client_model = ClientModel()
        self.channel = channel
        super().__init__()

    def on_modified(self, event):
        file_name = event.src_path.strip()
        if file_name.endswith("clients.json"):
            logging.debug("Reloading client config...")
            clients = self.client_model.get()
            updated_client_ids = list(clients.keys())
            self.update_clients(updated_client_ids)

    def update_clients(self, updated_client_ids):
        global RUNNING_CLIENTS
        logging.debug("Old: " + str(RUNNING_CLIENTS))
        logging.debug("New: " + str(updated_client_ids))

        for client_id in updated_client_ids:
            if client_id not in RUNNING_CLIENTS:
                run_client(client_id, self.channel)
                logging.info(f"Started client: {client_id}")

        for client_id in RUNNING_CLIENTS:
            if client_id not in updated_client_ids:
                self.stop_client(client_id)
                logging.info(f"Shutdown client: {client_id}")

    def stop_client(self, client_id):
        global RUNNING_CLIENTS
        global CONSUMER_TAGS
        stopping = False
        try:
            logging.info(f"Stopping client: {client_id}")
            client_tags = CONSUMER_TAGS[client_id]

            self.channel.basic_cancel(client_tags[f"{client_id}-broadcast"])
            self.channel.basic_cancel(client_tags[f"{client_id}-outbox"])
            self.channel.queue_unbind(
                queue=f"{client_id}-inbox",
                exchange=client_tags[f"{client_id}-inbox-publish"],
            )
            self.channel.queue_unbind(
                queue=f"{client_id}-inbox",
                exchange=client_tags[f"{client_id}-inbox-broadcast"],
            )

            if client_id in RUNNING_CLIENTS:
                RUNNING_CLIENTS.remove(client_id)
            stopping = True
        except Exception as error:
            logging.error(str(error))
        return stopping


class WatchConfigThread(threading.Thread):
    def __init__(self, running_clients, channel):
        threading.Thread.__init__(self)
        self.daemon = True
        self.running_clients = running_clients
        self.channel = channel
        self.start()

    def run(self):
        logging.info("Starting config watcher...")
        event_handler = ConfigHandler(self.channel)
        observer = Observer()
        observer.schedule(event_handler, path="./data", recursive=False)
        observer.start()

        while True:
            try:
                time.sleep(1)
            except KeyboardInterrupt as interrupt:
                observer.stop()
                raise interrupt


class SoarBusThread(threading.Thread):
    def __init__(self, clients, channel):
        threading.Thread.__init__(self)
        self.daemon = True
        self.clients = clients
        self.channel = channel
        self.start()

    def run(self):
        logging.info("Starting SOAR bus...")

        publish(self.channel, "soar-publish", EXCHANGES.get("publish"))

        for id in self.clients.keys():
            run_client(id, self.channel)


def run_client(client_id, channel):
    global RUNNING_CLIENTS
    global CONSUMER_TAGS
    client_model = ClientModel()
    client = client_model.find(client_id)

    try:
        client_id = client["client_id"]
        logging.info(f"Running client: {client_id}")

        forward_consumer_tag = forward(channel, f"{client_id}-outbox", "soar-publish")
        broadcast_consumer_tag = broadcast(
            channel, f"{client_id}-broadcast", EXCHANGES.get("broadcast")
        )
        subscribe(
            channel,
            f"{client_id}-inbox",
            EXCHANGES.get("publish"),
            client["subscription"],
        )
        subscribe(channel, f"{client_id}-inbox", EXCHANGES.get("broadcast"))

        CONSUMER_TAGS[client_id] = {
            f"{client_id}-broadcast": broadcast_consumer_tag,
            f"{client_id}-outbox": forward_consumer_tag,
            f"{client_id}-inbox-publish": EXCHANGES.get("publish"),
            f"{client_id}-inbox-broadcast": EXCHANGES.get("broadcast"),
        }

        if client_id not in RUNNING_CLIENTS:
            logging.debug(f"Appending client_id '{client_id}'")
            RUNNING_CLIENTS.append(client_id)

    except Exception as error:
        logging.error(str(error))


def on_channel_open(channel):
    # Invoked when the channel is open
    global RUNNING_CLIENTS

    client_model = ClientModel()
    clients = client_model.get()
    client_count = len(clients.keys())
    logging.debug(f"Running {client_count} clients")

    SoarBusThread(clients, channel)
    WatchConfigThread(RUNNING_CLIENTS, channel)


def on_connection_open(connection):
    # Invoked when the connection is open
    connection.channel(on_open_callback=on_channel_open)


def on_connection_close(connection, exception):
    # Invoked when the connection is closed
    connection.ioloop.stop()


if __name__ == "__main__":
    pingcounter = 0
    retries = 10
    interval = 5
    isreachable = False
    while isreachable is False and pingcounter < retries:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            s.connect((os.getenv("MQ_HOST", "localhost"), 5672))
            isreachable = True
        except socket.error as e:
            logging.info(f"RabbitMQ unavailable: retrying in {interval}s")
            time.sleep(interval)
            pingcounter += 1
        s.close()

    host = os.getenv("MQ_HOST", "localhost")
    connection = pika.SelectConnection(
        pika.ConnectionParameters(host),
        on_open_callback=on_connection_open,
        on_close_callback=on_connection_close,
    )

    try:
        connection.ioloop.start()
    except KeyboardInterrupt:
        # Loop until fully closed
        connection.close()
        connection.ioloop.start()
